/*
  Hatari - natfeats.h

  This file is distributed under the GNU General Public License, version 2
  or at your option any later version. Read the file gpl.txt for details.
*/

#ifndef HATARI_NATFEATS_H
#define HATARI_NATFEATS_H

extern bool NatFeat_ID(Uint32, Uint32 *retval);
extern bool NatFeat_Call(Uint32, bool isSuper, Uint32 *retval);

#endif /* HATARI_NATFEATS_H */
