/*
 * Hatari - fakeTosData.c
 *
 * This file is distributed under the GNU General Public License, version 2
 * or at your option any later version. Read the file gpl.txt for details.
 *
 * NOTE: This file is included by tos.c - do not compile and link this file
 * separately!
 */
const char FakeTosData_fileid[] = "Hatari fakeTosData.c";

/*
 * This is the assembled code from fakeTos.s, our ROM TOS replacement for
 * testing. NOTE: Remove first 0x1c (PRG_HEADER_SIZE) bytes from the assembled
 * program file or use an assembler like TurboAss or vasm that can generate
 * absolute binary images.
 *
 * Convert the assembled binary with:
 *
 *  hexdump -v -e ' 16/1 "0x%02x," "\n" ' fakeTos.bin > fakeTos.txt
 *
 * Then replace the data below in FakeTos_data[] with the contents of
 * fakeTos.txt (edit the last line to remove empty "0x,")
 */
const Uint8 FakeTos_data[] =
{
0x60,0x2e,0x00,0x01,0x00,0xe0,0x00,0x30,0x00,0xe0,0x00,0x00,0x00,0x00,0x10,0x00,
0x00,0xe0,0x00,0x30,0x00,0x00,0x00,0x00,0x03,0x03,0x20,0x18,0x00,0x01,0x4c,0x63,
0x00,0x00,0x08,0x80,0x00,0x00,0x08,0x70,0x00,0x00,0x08,0x00,0x00,0x00,0x00,0x00,
0x46,0xfc,0x27,0x00,0x4e,0x70,0x11,0xfc,0x00,0x05,0x80,0x01,0x4f,0xf9,0x00,0x02,
0x00,0x00,0x43,0xfa,0x00,0x7e,0x30,0x7c,0x00,0x08,0x20,0xc9,0xb0,0xfc,0x01,0xc0,
0x6f,0xf8,0x43,0xfa,0x00,0x84,0x23,0xc9,0x00,0x00,0x00,0x68,0x23,0xc9,0x00,0x00,
0x00,0x72,0x41,0xf8,0xfa,0x00,0x11,0x7c,0x00,0x48,0x00,0x11,0x41,0xf8,0xfc,0x00,
0x10,0xbc,0x00,0x03,0x10,0xbc,0x00,0x16,0x41,0xf9,0x00,0xfa,0x00,0x00,0x0c,0x90,
0xab,0xcd,0xef,0x42,0x66,0x02,0x00,0x0a,0x70,0x00,0x20,0x40,0x22,0x40,0x46,0xfc,
0x07,0x00,0x4f,0xf9,0x00,0x01,0x80,0x00,0x48,0x78,0x10,0x00,0x48,0x7a,0xff,0x62,
0x4e,0xf8,0x11,0x00,0x45,0x52,0x52,0x4f,0x52,0x3a,0x20,0x55,0x6e,0x68,0x61,0x6e,
0x64,0x6c,0x65,0x64,0x20,0x65,0x78,0x63,0x65,0x70,0x74,0x69,0x6f,0x6e,0x21,0x0d,
0x0a,0x00,0x48,0x7a,0xff,0xe0,0x3f,0x3c,0x00,0x09,0x4e,0x41,0x5c,0x8f,0x3f,0x3c,
0x00,0x01,0x3f,0x3c,0x00,0x4c,0x4e,0x41,0x4e,0x73
};
