/*
  Hatari - microphone.h

  This file is distributed under the GNU General Public License, version 2
  or at your option any later version. Read the file gpl.txt for details.
*/

#ifndef HATARI_MICROPHONE_H
#define HATARI_MICROPHONE_H

extern bool Microphone_Start (int sampleRate);
extern void Microphone_Stop (void);

#endif /* HATARI_MICROPHONE_H */
