/*
  Hatari - version.h

  Name and version for window title (and other outputs)
*/

/* devel */
//#define PROG_NAME "Hatari v2.4.0-devel (" __DATE__ ")"

/* release */
#define PROG_NAME "Hatari v2.4.1"
